
import java.io.BufferedReader;
import java.io.FileInputStream;
import java.io.FileReader;
import java.io.FileWriter;
import java.io.IOException;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.HashMap;
import java.util.LinkedHashSet;
import java.util.List;
import java.util.Map;
import java.util.Properties;
import java.util.Set;

import org.biojava.bio.symbol.*;
import org.biojava.bio.seq.*;

public class CompareMRNA_MiRNATail {

	public static String commonPath;
	public static Properties properties;
	
	
	Map<String, String> parsedMRNA = new HashMap<String,String>();
	Map<String, String> parsedMiRNA = new HashMap<String,String>();
	
	public double threshold;
	public double cut;
	public int offset;
	private BufferedReader mRNAmiRNA_head;
	String mRNAFile;
	String miRNAFile;
	
	
	
	
	public CompareMRNA_MiRNATail(String _mRNAFile, String _miRNAFile, int _offset, double _threshold) {
		this.mRNAFile= _mRNAFile;
		this.miRNAFile= _miRNAFile;
		this.offset= _offset;
		this.threshold= _threshold;
		//this.cut= _cut;
	}
	

	public static void init() throws IOException {
		properties = new Properties();
		properties.load(new FileInputStream("prefs.properties"));
		commonPath = properties.getProperty("commonPath");
	}

	
	private void start() {
		System.out.println("Start");
		
		//String inputFile="val_mRNA_hsa_4.fas";
		
		parsedMRNA = parseMRNA(commonPath+mRNAFile);
		System.out.println( "parsedMRNA.size()= "+parsedMRNA.size() );
		parsedMiRNA = parseMiRNA(commonPath+miRNAFile);
		System.out.println( "parsedMiRNA.size()= "+parsedMiRNA.size() );

		int[][] miRNATailPWM;
		int[][] targetTailPWM;
		
		int rowcount =0;
		String row = null;
		try {
			mRNAmiRNA_head = new BufferedReader(new FileReader(commonPath+"mRNA-miRNA_CalculatorHead_fragment_"+mRNAFile+".txt"));
			FileWriter outputTailTargets = new FileWriter(commonPath+"TailConfirmedTargets_"+mRNAFile+"_th"+threshold+"_offset"+offset+".txt"); 

			//leggo le singole celle (una per riga) contenenti i mRNA proiettati sulla mappa addestrata
			while ((row  = mRNAmiRNA_head.readLine()) != null) {
				rowcount++;
				System.out.println("row= "+rowcount); 
				String[] rowParts = row.split(";");
				if (rowParts.length != 3 || rowParts[1].isEmpty() || rowParts[2].isEmpty()) continue;
				String[] targets = rowParts[1].split(",");
				String[] miRNAs = rowParts[2].split(",");
				//System.out.println(targets.length + " target.");
				//System.out.println(miRNAs.length + " mirna.");
				//qui conservo i target ed i mirna che sopravvivono
				ArrayList<String> mirnaList= new ArrayList<String>();
				ArrayList<String> targetList= new ArrayList<String>();
				
				//per ogni mirna presente nella cella, vado a calcolare 
				//la distanza tra la sua coda e quella del potenziale target
				for (String mirna : miRNAs) {

					String mirnaTail= parsedMiRNA.get(mirna);
//					System.out.println(mirna +": "+mirnaTail);
					//prendo in considerazione solo la coda del mirna
					mirnaTail= mirnaTail.substring(8, mirnaTail.length());
					//poiche le code possono avere lunghezza differente, segno la lunghezza
					int tailLength= mirnaTail.length();

					//System.out.println(mirna +": "+mirnaTail);

					for  (String target : targets) {

						String [] splitTarget = target.split("_");
						
						//System.out.println("splitTarget[0]: "+ splitTarget[0]);
						//System.out.println("splitTarget[1]: "+ splitTarget[1]);
						
						//verifico la sequenza introducendo + o - offset, che tiene in considerazione il bulge
						for (int off= -offset; off <= offset; off++){
							
							
							int startPositionMRNATail = Integer.parseInt(splitTarget[1]) - tailLength + off;
							int endPositionMRNATail = Integer.parseInt(splitTarget[1]) - 1 + off;

							if (startPositionMRNATail < 0) continue;

							//System.out.println("start: "+startPositionMRNATail);
							//System.out.println(("end: "+endPositionMRNATail));

							String targetTail= parsedMRNA.get(splitTarget[0]);
							targetTail= targetTail.substring(startPositionMRNATail, endPositionMRNATail);

							//System.out.println("target: "+targetTail); 

							//considero il reverse complement di mirnaTail per poterlo confrontare con targetTail
							String mirnaTailRC= ReverseComplement(mirnaTail);

							//converto sia mirnaTailRC che targetTail in PWM
							miRNATailPWM = new int[tailLength][4];
							targetTailPWM = new int[tailLength][4];

							changeToBits(miRNATailPWM.length,miRNATailPWM,mirnaTailRC);
							changeToBits(targetTailPWM.length,targetTailPWM,targetTail);

							//calcolo la distanza tra le PWM
							double distance= calculateNEDistance(miRNATailPWM, targetTailPWM);
//							double distance= calculateApproxKLDistance(miRNATailPWM, targetTailPWM);
							//System.out.println(distance);

							//controllo la soglia sulla distanza
							if (distance <= threshold){
								targetList.add(target);
								//System.out.println(rowParts[0]+";"+target+";"+mirna);
							}
							
						}
						
					} //end for target

					mirnaList.add(mirna);
					
				} //end for mirna
				
				
				//Crea la riga del file contentnete le coordiante del neurone, 
				//la lista dei MRNA e la lista dei MiRNA (SE CI SONO)
				if (!targetList.isEmpty() && !mirnaList.isEmpty()){
					//sbobino la lista dei mirna
					String[] mL= mirnaList.toArray(new String[mirnaList.size()]);
					String ms="";
					for (int i=0; i < mirnaList.size(); i++){
						if (i==0){
							ms= ms + mL[0];
						}
						else{
							ms= ms +"," + mL[i];
						}
					}
					//sbobino la lista dei target
					//elimino i duplicati con un LinkedHashSet
					Set<String> tSet = new LinkedHashSet<String>(targetList);
					targetList.clear();
					targetList.addAll(tSet);
					//System.out.println("dopo la cura: "+targetList.size()); 
					String[] tL= targetList.toArray(new String[targetList.size()]);
					String ts="";
					for (int i=0; i < targetList.size(); i++){
						if (i==0){
							ts= ts + tL[0];
						}
						else{
							ts= ts +"," + tL[i];
						}
					}
					
					//System.out.println(rowParts[0]+";"+ ts +";"+ms); 
					outputTailTargets.write(rowParts[0]+";"+ ts +";"+ms+"\n"); 
				}

			} // end while row

			outputTailTargets.close();
			

		} catch (Exception e) {
			e.printStackTrace();
		}
		
	}
	
	
	
	
	public static void main(String[] args) throws IOException {
		
		init();
		if (args.length != 4) {
			System.out.println("Input example: java CompareMRNA_MiRNATail_hsa.java mRNAFile miRNAFile offset threshold");
			return;
		}
		
		String mRNAFile= args[0];   //inputFile="val_mRNA_hsa_4.fas";
		String miRNAFile= args[1];
		int offset = Integer.parseInt(args[2]); //3
		double threshold = Double.parseDouble(args[3]);  // 0.7
		//double cut = Double.parseDouble(args[4]);  //0.57



			CompareMRNA_MiRNATail compareMRNA_MiRNA = new CompareMRNA_MiRNATail(mRNAFile, miRNAFile, offset, threshold);

			compareMRNA_MiRNA.start();

	}
	
	
	

	
	
	
	public static HashMap<String, String> parseMRNA(String file) {
		HashMap<String,String> result = new HashMap<String,String>();

		BufferedReader seqInput = null;
		String row = "";	
		String sequence = "";
		String MRNA = "";
		try {
			seqInput = new BufferedReader(new FileReader(file));
			while((row = seqInput.readLine())!=null){
				if ((row.indexOf(">")) == -1){
					sequence = sequence + row;
				}
				else {
					if(MRNA.length()!=0) {
						result.put(MRNA, sequence);
						sequence = "";						
					}
					MRNA = row;
				}
			}
			result.put(MRNA, sequence); // aggiunge l'ultima coppia
			seqInput.close();
		} catch (IOException e) {
			throw new RuntimeException("Errore nella lettura!", e);
		}

		return result;
	}

	public static HashMap<String, String> parseMiRNA(String file) {
		HashMap<String,String> result = new HashMap<String,String>();

		BufferedReader seqInput = null;
		String row = "";	
		String sequence = "";
		String MIRNA = "";
		try {
			seqInput = new BufferedReader(new FileReader(file));
			while((row = seqInput.readLine())!=null){
				if ((row.indexOf(">")) == -1){
					sequence = sequence + row;
				}
				else {
					if(MIRNA.length()!=0) {
						result.put(MIRNA, sequence);
						sequence = "";						
					}						
					String[] realKey = row.split(" ");
					MIRNA = realKey[0];
				}
			}
			result.put(MIRNA, sequence); // aggiunge l'ultima coppia
			seqInput.close();
		} catch (IOException e) {
			throw new RuntimeException("Errore nella lettura!", e);
		}

		return result;
	}
	
	
	public String ReverseComplement(String str) {
		   String retval = null;
			try {
		      //make a RNA SymbolList
		      SymbolList symL = RNATools.createRNA(str);
		 
		      //reverse complement it
		      symL = RNATools.reverseComplement(symL);
		 
		      //prove that it worked
		      retval = symL.seqString();
		    }
		    catch (IllegalSymbolException ex) {
		      //this will happen if you try and make the RNA seq using non IUB symbols
		      ex.printStackTrace();
		    }catch (IllegalAlphabetException ex) {
		      //this will happen if you try and reverse complement a non RNA sequence using RNATools
		      ex.printStackTrace();
		    }
			
			return retval;
	}
		
	/**
	 * Converte una sequenza di nucleotidi in una PWM
	 * 
	 * @param sL - lunghezza della sequenza
	 * @param trv - matrice
	 * @param sequence - sequenza DNA
	 */
	private void changeToBits(int sL, int[][] trv, String sequence){
		int spit,plus;	
		char nt;
		int pts,j,p;
		
		sequence=sequence.toUpperCase();
		spit=sequence.length();

		if (spit<sL) {
			pts=sL-spit;
			plus=0;
		} else {
			pts=0;
			plus=spit-sL;
		}
		
		for (j=0;j<pts;j++){
			for (p=0;p<4;p++){
				trv[j][p]=0;
			}
		}
		
		for (j=pts;j<sL;j++){
			nt=sequence.charAt(j-pts+plus);
			switch (nt) {
			case 'A': trv[j][0]=1;
				break;
			case 'C': trv[j][1]=1;
				break;
			case 'G': trv[j][2]=1;
				break;
			case 'U': trv[j][3]=1;
				break;
			case 'T': trv[j][3]=1;
				break;
			default:  break;
			}
		}
	}
	
	
	//Normalized Euclidean Distance for PWM
	public double calculateNEDistance(int pwm1[][], int pwm2[][]) {
		double nt_diff=0;
		double[][] diff = new double[pwm1.length][4];
		double diffSum = 0;
		for (int i=0; i<pwm1.length; i++){
			double tempSum=0.0;
			for (int j=0; j<4; j++){
				nt_diff=0;
				diff[i][j]=pwm1[i][j]-pwm2[i][j];
				nt_diff=diff[i][j]*diff[i][j];
				tempSum = tempSum + nt_diff;
			}
			diffSum= diffSum + Math.sqrt(tempSum);
		}
		diffSum= diffSum * 1.0/(Math.sqrt(2)*pwm1.length);
		return diffSum;
	}
	
	
	/**
	 * Distance used for Motif comparison 
	 * @param input - Position Weight Matrix (probability described by a frequency) representation for a DNA sequence
	 * @return distance
	 */
	public double calculateApproxKLDistance(int input[][], int input2[][]) {
		
		double[][] diff = new double[input.length][4];
		double diffSum = 0.0;
		double pwm1= 0.0;
		double pwm2= 0.0;
		for (int i=0; i<input.length; i++){
			double kl_dist=0;
			//for each nucleotide A,C,G,T
			for (int j=0; j<4; j++){
				
				if (input[i][j] != 0.0) 
					pwm1= input[i][j];
				else  //evitiamo eccezione divisione per zero
					pwm1= 0.0000000000000001;
				if (input2[i][j]  != 0.0)
					pwm2= input2[i][j]; 
				else //evitiamo eccezione divisione per zero
					pwm2= 0.0000000000000001;
				//calculate KL(1,2)j
				double kl12= pwm1 * Math.log( pwm1 / pwm2);			
				//calculate KL(2,1)j
				double kl21= pwm2 * Math.log( pwm2 / pwm1);
				
				//sum results
				kl_dist = kl_dist + kl12 + kl21;
			}
			diffSum = diffSum + kl_dist;
		}
		diffSum= diffSum/ (2.0 * input.length);
			
		
		return diffSum;
	}
	
	
	
	
}
