

import java.util.Random ;
import java.io.*;
import java.util.*;

// @ Liisa Heikkinen 2010

public class Neuron {
	public int input_dimension;
	public double nodeWeight[][];	
	public double diff[][];
	public double diffSum;

	public Neuron() { 
		this(0);
	}

	public Neuron(int n) { 
		input_dimension = n;
		diffSum = 0;
	}

	public void allocWeights(int n, int x, int y) {
		input_dimension = n;
		nodeWeight = new double[input_dimension][4];		
	}
	
	public void setupWeights(int n, int x, int y) {
		input_dimension = n;
		nodeWeight = new double[input_dimension][4];
		int arpa;
		for (int i=0; i<input_dimension; i++){
			arpa = (int)(Math.random()*4);
			nodeWeight[i][arpa] = 1; 
		}
	}

	public void printNodeWeight(PrintWriter outFile, int x, int y) {
		String row;
		outFile.println(x+":"+y);	
		for (int i=0; i<4; i++){
			row=""+nodeWeight[0][i];
			for (int j=1; j<input_dimension; j++){
				row=row+" "+nodeWeight[j][i];
			}
			outFile.println(row);
		}
	}

	public double calculateDistance(int input[][]) {
		double nt_diff=0;
		diff = new double[input_dimension][4];
		diffSum = 0;
		for (int i=0; i<input_dimension; i++){
			for (int j=0; j<4; j++){
				nt_diff=0;
				diff[i][j]=input[i][j]-nodeWeight[i][j];
				nt_diff=diff[i][j]*diff[i][j];
				diffSum = diffSum + nt_diff;
			}
		}
		return diffSum;
	}

		
	//Normalized Euclidean Distance for PWM
	public double calculateNEDistance(int input[][]) {
		double nt_diff=0;
		diff = new double[input.length][4];
		diffSum = 0;
		for (int i=0; i<input.length; i++){
			double tempSum=0.0;
			for (int j=0; j<4; j++){
				nt_diff=0;
				diff[i][j]=input[i][j]-nodeWeight[i][j];
				nt_diff=diff[i][j]*diff[i][j];
				tempSum = tempSum + nt_diff;
			}
			diffSum= diffSum + Math.sqrt(tempSum);
		}
		diffSum= diffSum * 1.0/(Math.sqrt(2)*input.length);
		return diffSum;
	}
	
	/**
	 * Tiene conto dello shift tra mirna di lunghezze diverse.
	 * nodeWeight e il max length vector
	 * @param input
	 * @return
	 */
	//Normalized Euclidean Distance for PWM
		public double calculateNEDistanceTail(int input[][]) {
			double nt_diff=0;
			diff = new double[input.length][4];
			diffSum = 0;
			//for (int i= nodeWeight.length; i >= nodeWeight.length - input.length; i--){
			for (int i=0; i<input.length; i++){
				double tempSum=0.0;
				for (int j=0; j<4; j++){
					nt_diff=0;
					diff[i][j]=input[i][j]-nodeWeight[i+(nodeWeight.length - input.length)][j];
					nt_diff=diff[i][j]*diff[i][j];
					tempSum = tempSum + nt_diff;
				}
				diffSum= diffSum + Math.sqrt(tempSum);
			}
			diffSum= diffSum * 1.0/(Math.sqrt(2)*input.length);
			return diffSum;
		}

	
		/**
		 * Tiene conto dello shift tra mirna di lunghezze diverse.
		 * nodeWeight e il min length vector
		 * @param input
		 * @return
		 */
		//Normalized Euclidean Distance for PWM
		public double calculateNEDistanceTail2(int input[][]) {
					double nt_diff=0;
					diff = new double[nodeWeight.length][4];
					diffSum = 0;
					//for (int i= nodeWeight.length; i >= nodeWeight.length - input.length; i--){
					for (int i=0; i<nodeWeight.length; i++){
						double tempSum=0.0;
						for (int j=0; j<4; j++){
							nt_diff=0;
							diff[i][j]=input[i][j]-nodeWeight[i][j];
							nt_diff=diff[i][j]*diff[i][j];
							tempSum = tempSum + nt_diff;
						}
						diffSum= diffSum + Math.sqrt(tempSum);
					}
					diffSum= diffSum * 1.0/(Math.sqrt(2)*input.length);
					return diffSum;
				}
		
		
	
	/**
	 * Distance used for Motif comparison 
	 * @param input - Position Weight Matrix (probability described by a frequency) representation for a DNA sequence
	 * @return distance
	 */
	public double calculateApproxKLDistance(int input[][]) {
		
		diff = new double[input_dimension][4];
		double diffSum = 0.0;
		double pwm1= 0.0;
		double pwm2= 0.0;
		for (int i=0; i<input_dimension; i++){
			double kl_dist=0;
			//for each nucleotide A,C,G,T
			for (int j=0; j<4; j++){
				
				if (input[i][j] != 0.0) 
					pwm1= input[i][j];
				else  //evitiamo eccezione divisione per zero
					pwm1= 0.0000000000000001;
				if (nodeWeight[i][j]  != 0.0)
					pwm2= nodeWeight[i][j]; 
				else //evitiamo eccezione divisione per zero
					pwm2= 0.0000000000000001;
				//calculate KL(1,2)j
				double kl12= pwm1 * Math.log( pwm1 / pwm2);			
				//calculate KL(2,1)j
				double kl21= pwm2 * Math.log( pwm2 / pwm1);
				
				//sum results
				kl_dist = kl_dist + kl12 + kl21;
			}
			diffSum = diffSum + kl_dist;
		}
		diffSum= diffSum/ (2.0 * input_dimension);
			
		
		return diffSum;
	}
	
	/**
	 * Distance used for Motif comparison 
	 * @param input - Position Weight Matrix (probability described by a frequency) representation for a DNA sequence
	 * @return distance
	 */
	public double calculateKLDistance(int input[][]) {
		
		diff = new double[input_dimension][4];
		double diffSum = 0.0;
		for (int i=0; i<input_dimension; i++){
			double kl_dist=0;
			//for each nucleotide A,C,G,T
			for (int j=0; j<4; j++){
				//calculate KL(1,2)j
				double kl12;
				if (input[i][j] == 0.0){ //evitiamo eccezione divisione per zero
					kl12 = 0;
				}
				else if (nodeWeight[i][j]  == 0.0){ //evitiamo eccezione valore infinito
					kl12 = Double.MAX_VALUE;
				}
				else{	
					kl12= Math.log( input[i][j] / nodeWeight[i][j]);
					kl12= 1;
				}
				//calculate KL(2,1)j
				double kl21;
				if (nodeWeight[i][j]  == 0.0){  //evitiamo eccezione divisione per zero
					kl21 = 0;
				}
				else if (input[i][j] == 0.0){  //evitiamo eccezione valore infinito
					kl21 = 1;
				}
				else {			
					kl21= Math.log( nodeWeight[i][j] / input[i][j] );
					kl21= kl21 * nodeWeight[i][j];
				}

				kl_dist = kl_dist + kl12 + kl21;
			}
			diffSum = diffSum + kl_dist;
		}
		diffSum= diffSum/2.0;
			
		
		return diffSum;
	}
	
	
//	public double calculateDistanceTail(int input[][]) {
//		double nt_diff=0;
//		diff = new double[input.length][4];
//		diffSum = 0;
//		for (int i=0; i<input.length; i++){
//			for (int j=0; j<4; j++){
//				nt_diff=0;
//				diff[i][j]=input[i][j]-nodeWeight[i][j];
//				nt_diff=diff[i][j]*diff[i][j];
//				diffSum = diffSum + nt_diff;
//			}
//		}
//		return diffSum;
//	}
//
//	
//	public double calculateDistanceTail2(int input[][]) {
//		double nt_diff=0;
//		diff = new double[input.length][4];
//		diffSum = 0;
//		for (int i=input_dimension-input.length; i<input_dimension; i++){
//			for (int j=0; j<4; j++){
//				nt_diff=0;
//				diff[i][j]=input[i][j]-nodeWeight[i][j];
//				nt_diff=diff[i][j]*diff[i][j];
//				diffSum = diffSum + nt_diff;
//			}
//		}
//		return diffSum;
//	}
	

	
/*	
  public double calculateDistance(int input[][]) {
		double nt_diff=0;
		diff = new double[input_dimension][4];
		diffSum=0;
		int mask=1;
		for (int i=0; i<input_dimension; i++){
			if (i==input_dimension-8 || i==input_dimension-1) mask=2;
			if (i==input_dimension-7) mask=5;
			
			for (int j=0; j<4; j++){
				nt_diff=0;
				diff[i][j]=input[i][j]-nodeWeight[i][j];
				nt_diff=mask*diff[i][j]*diff[i][j];
				diffSum = diffSum + nt_diff;
			}
		}
		return diffSum;
	}
*/

	public void updateWeights(int input[][],double dist, double alpha, double r2) {
		double h=alpha*Math.exp(-dist/(2*r2)); 
		for (int i=0; i<input.length; i++){
			for (int j=0; j<4; j++){
				nodeWeight[i][j]=nodeWeight[i][j]+h*diff[i][j];
			}
		}
	}
	
	public void updateWeightsTail(int input[][],double dist, double alpha, double r2) {
		double h=alpha*Math.exp(-dist/(2*r2)); 
		for (int i=0; i<input.length; i++){
		//for (int i= nodeWeight.length; i >= nodeWeight.length - input.length; i--){
			for (int j=0; j<4; j++){
				nodeWeight[i+(nodeWeight.length - input.length)][j]=nodeWeight[i+(nodeWeight.length - input.length)][j]+h*diff[i][j];
			}
		}
	}
	
	
	public void updateWeightsTail2(int input[][],double dist, double alpha, double r2) {
		double h=alpha*Math.exp(-dist/(2*r2)); 
		for (int i=0; i<nodeWeight.length; i++){
		//for (int i= nodeWeight.length; i >= nodeWeight.length - input.length; i--){
			for (int j=0; j<4; j++){
				nodeWeight[i][j]=nodeWeight[i][j]+h*diff[i][j];
			}
		}
	}

} // class
