
import java.io.*;
import java.util.*;


public class SOMProjectionMRNA{

	public static String commonPath;
	public static Properties properties;


	public class SOM {
		private ArrayList<String> [][] mapmRNA;  
		private TreeSet<String> [][] mapmRNA2;  
		private ArrayList<String> [][] mapmiRNA;
		
		public ArrayList<String>[][] getMapmRNA() {
			return mapmRNA;
		}
		public void setMapmRNA(ArrayList<String>[][] mapmRNA) {
			this.mapmRNA = mapmRNA;
		}
		
		public TreeSet<String>[][] getMapmRNA2() {
			return mapmRNA2;
		}
		public void setMapmRNA2(TreeSet<String>[][] mapmRNA2) {
			this.mapmRNA2 = mapmRNA2;
		}

		public ArrayList<String>[][] getMapmiRNA() {
			return mapmiRNA;
		}
		public void setMapmiRNA(ArrayList<String>[][] mapmiRNA) {
			this.mapmiRNA = mapmiRNA;
		}
	}

	public static void init() throws IOException {
		properties = new Properties();
		properties.load(new FileInputStream("prefs.properties"));
		commonPath = properties.getProperty("commonPath");
	}

	
	public static void main(String [] args) throws IOException{
		init();
		if (args.length != 5) {
			System.out.println("Input example: java CalculatorHead trainedSOMFile mRNAFile miRNAFile 30 5");
			return;
		}

		String trainedSOMFile= args[0];
		String mRNAFile= args[1];
		String miRNAFile= args[2];
		String mapSize= args[3];
		String thresholdHead= args[4];
	
		SOM som;
	
		try {
			som= new SOMProjectionMRNA().computeSOM(trainedSOMFile, mRNAFile, miRNAFile, Integer.parseInt(mapSize), Double.parseDouble(thresholdHead));
		} catch (RuntimeException e) {
			String msg = e.getMessage();
			switch (msg) {
			case "File not found": 
			case "Errore nella lettura!": 
			case "Errore durante la chiusura del file!": 
				System.out.println(msg);
				return;
			default: throw e;
			}
		}
				
		ArrayList<String>[][] mapmRNA = som.getMapmRNA();
		TreeSet<String>[][] mapmRNA2 = som.getMapmRNA2();
		ArrayList<String>[][] mapmiRNA = som.getMapmiRNA();
		PrintWriter outputFile = null;	
		PrintWriter outputFile2 = null;	

		try {
			outputFile = new PrintWriter(new FileOutputStream(commonPath+"mRNA-miRNA_CalculatorHead_fragment_"+mRNAFile+".txt"),true);
			outputFile2 = new PrintWriter(new FileOutputStream(commonPath+"mRNA-miRNA_CalculatorHead_"+mRNAFile+".txt"),true);

		} catch (IOException e1) {
			System.out.println("Non e' possibile scrivere!");
		}		

	    // outpuFile
		for (int v=0; v<mapmRNA.length; v++) {
			for (int w=0; w<mapmRNA[v].length; w++) {
				printAndSave((v+","+w+";"), outputFile);
				for (Iterator<String> iter = mapmRNA[v][w].iterator(); iter.hasNext(); ) {
					printAndSave(iter.next(), outputFile);
					if(iter.hasNext()) printAndSave((","), outputFile);
				}
				printAndSave((";"), outputFile);
				for (Iterator<String> iter = mapmiRNA[v][w].iterator(); iter.hasNext(); ) {
					printAndSave(iter.next(), outputFile);
					if(iter.hasNext()) printAndSave((","), outputFile);
				}	
				printAndSave(("\r\n"), outputFile);
			}
		}
		
	    // outpuFile2
		for (int v=0; v<mapmRNA2.length; v++) {
			for (int w=0; w<mapmRNA2[v].length; w++) {
				printAndSave((v+","+w+";"), outputFile2);
				for (Iterator<String> iter = mapmRNA2[v][w].iterator(); iter.hasNext(); ) {
					printAndSave(iter.next(), outputFile2);
					if(iter.hasNext()) printAndSave((","), outputFile2);
 				}
				printAndSave((";"), outputFile2);
				for (Iterator<String> iter = mapmiRNA[v][w].iterator(); iter.hasNext(); ) {
					printAndSave(iter.next(), outputFile2);
					if(iter.hasNext()) printAndSave((","), outputFile2);
				}	
				printAndSave(("\r\n"), outputFile2);
			}
		}
		
		if (outputFile != null)
			outputFile.close(); 

		if (outputFile2 != null)
			outputFile2.close(); 
		
	} // end main
	
	
	public SOM computeSOM (String filetrainedSOM, String filemRNA, String filemiRNA, int mapSize, double thresholdHead){
		
		BufferedReader seqInput = null;

		int epochs=1;
		String TrainedSOMfile= filetrainedSOM;
		
		File ListmRNA;
		File ListmiRNA;
		
		ArrayList<String> seqmRNA = new ArrayList<String>();  // Contiene le sequenze dei geni
		ArrayList<String> seqmiRNA = new ArrayList<String>();  // Contiene le sequenze dei miRNA

		ArrayList<String> seq2mRNA = new ArrayList<String>();  // Contiene la lista delle finestre per un dato gene d'ingresso 
		ArrayList<Integer> seq2mRNAidx = new ArrayList<Integer>();  // Indice del gene 
		ArrayList<Integer> seq2mRNAsubidx = new ArrayList<Integer>();  // Sottoindice del frammento all'interno del gene 

		ArrayList<String> annotSeqmRNA = new ArrayList<String>(); // Contiene le annotazioni dei geni 
		ArrayList<String> annotSeqmiRNA = new ArrayList<String>(); // Contiene le annotazioni dei miRNA 

		ArrayList<String> [][] mapmRNA;  // Mappa addestrata dei geni (o mRNA)
		TreeSet<String> [][] mapmRNA2; 
		ArrayList<String> [][] mapmiRNA; // Mappa addestrata dei miRNA
		
		
		ListmRNA = new File(filemRNA);
		ListmiRNA = new File(filemiRNA);
		
		int size = mapSize;
		int seqLength = 8; // mirNA's SEED
		String seqApp = "";
		String row = "";
		
		mapmRNA = new ArrayList[size][size];
		mapmRNA2 = new TreeSet[size][size];
		mapmiRNA = new ArrayList[size][size];
		
		for (int i=0; i<size; i++) {
			for (int j=0; j<size; j++) {
				mapmRNA[i][j] = new ArrayList<String>(); 
				mapmRNA2[i][j] = new TreeSet<String>(); 
				mapmiRNA[i][j] = new ArrayList<String>(); 
			}
		}
			

		if (!ListmRNA.exists()) {
			throw new RuntimeException("File not found");
		} else {
			try {
				seqInput = new BufferedReader(new FileReader(ListmRNA));
				while((row = seqInput.readLine())!=null){
					if ((row.indexOf(">"))!=-1){
						annotSeqmRNA.add(row);
						if (seqApp.length() > 0) {
							seqmRNA.add(seqApp);
							seqApp = "";
						}
					} else seqApp = seqApp + row; 
				}
				if (seqApp.length() > 0) seqmRNA.add(seqApp);
			} catch (IOException e1) {
				throw new RuntimeException("Errore nella lettura!", e1);
			}
			try {
				if (seqInput != null)
					seqInput.close();
			} catch (IOException e2) {
				throw new RuntimeException("Errore durante la chiusura del file!", e2);
			}
		}

		
		if (!ListmiRNA.exists()) {
			throw new RuntimeException("File not found");
		} else {
			try {
				seqInput = new BufferedReader(new FileReader(ListmiRNA));
				while((row = seqInput.readLine())!=null){
					if ((row.indexOf(">")) == -1){
						seqmiRNA.add(row);
					}
					else annotSeqmiRNA.add(row);
				}
			} catch (IOException e1) {
				throw new RuntimeException("Errore nella lettura!", e1);
			}
			try {
				if (seqInput != null)
					seqInput.close();
			} catch (IOException e2) {
				throw new RuntimeException("Errore durante la chiusura del file!", e2);
			}
		}


		for (int i=0; i<seqmRNA.size(); i++) {
			String fullSeq = seqmRNA.get(i);
			
			for(int j=0;j<fullSeq.length()-seqLength;j++) {
				String subSeq = fullSeq.substring(j,j+seqLength);
				seq2mRNA.add(subSeq);
			   	seq2mRNAidx.add(i);
			   	seq2mRNAsubidx.add(j);
			}
		}
	    
		int seqmiRNANumber = seqmiRNA.size();
		int seqmRNANumber = seqmRNA.size();
		int seq2mRNANumber = seq2mRNA.size();			

		int units=size*size;
	    int nr1,nr2;
			
	    double[][] Dist = new double[units][units];
		 for (int j1=0;j1<size;j1++){
			for (int i1=0;i1<size;i1++){	
				nr1=j1*size+i1;	
				for (int j2=0;j2<size;j2++){
					for (int i2=0;i2<size;i2++){
						nr2=j2*size+i2;
						if (nr1>nr2) 
							Dist[nr1][nr2]=Dist[nr2][nr1];
						else {
							Dist[nr1][nr2]=(j2-j1)*(j2-j1)+(i2-i1)*(i2-i1);
						}	
					}	
				}   
			}	
		 }
			

		// PASSO 1 DELL'ALGORITMO: Assegna ai vettori dei pesi valori casuali (INIZIALIZZAZIONE MAPPA)...
		Neuron[][] map = new Neuron[size][size];
		for (int i=0; i<size; i++){
			for (int j=0; j<size; j++) {
				map[i][j] = new Neuron();	
				map[i][j].allocWeights(seqLength,i,j);
			}
		}
			
		BufferedReader reader;
		String line = null;
		ArrayList<String> fileList = new ArrayList<String>();
		int contaRighe = 0;
		int mapI = -1;
		int mapJ = -1;

		try {
			reader = new BufferedReader(new FileReader(commonPath+TrainedSOMfile));

			while ((line = reader.readLine()) != null) {
				if ( (contaRighe % 5) == 0 ) {													
					if(contaRighe > 0) {
						for(int i1=0; i1 < fileList.size(); i1++) {
							String rigaPesi = fileList.get(i1);
							String[] pesiString = rigaPesi.split(" "); 
							for(int i2=0; i2<pesiString.length;i2++) {
								map[mapI][mapJ].nodeWeight[i2][i1] = Double.parseDouble(pesiString[i2]);
							}
						}

						fileList.clear();	
					}

					String[] mapIJ = line.split(":");
					mapI = Integer.parseInt(mapIJ[0]);
					mapJ = Integer.parseInt(mapIJ[1]);
				}
				else {
					fileList.add(line);
				}					    
				contaRighe++;
			}

			reader.close();

			for(int i1=0; i1 < fileList.size(); i1++) {
				String rigaPesi = fileList.get(i1);
				String[] pesiString = rigaPesi.split(" "); 
				for(int i2=0; i2<pesiString.length;i2++) {
					map[mapI][mapJ].nodeWeight[i2][i1] = Double.parseDouble(pesiString[i2]);
				}
			}

			fileList.clear();
		} catch (FileNotFoundException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}catch (IOException e) {
				// TODO Auto-generated catch block
				e.printStackTrace();
		}
				
		double distance;
		int min_i=-1;
		int min_j=-1;
		int min_i2=-1;
		int min_j2=-1;
		String trainingSample;
		
		double nRad=size;		// neighbor radius
		double r2=nRad*nRad+nRad*nRad;	
		double alpha=0.85;
		double alphastart = 0.85;	// training rate to start with
		double alphaend = 0.1;	// training rate in the end

		double qe=0;	// quantization error: Average distance between each input string and its best matching unit in the net
		double te=0;	// topographic error: the proportion of all input strings for which the best and the second best matching units are not adjacent
		int[][] trv;
		int ten=0;
		int pst=0;

		int updateStep=((epochs-2)*seqmRNANumber)/151;
		int updateStep2=((epochs-2)*seqmiRNANumber)/151;

	    double maxThreshold = 5;

// ELABORAZIONE sui mRNA
	    for (int k=0; k<epochs; k++){
	    	qe=0;	
	    	te=0;
	    	int[] order = new int[seqmRNANumber];
	    	changeorder(order);

	    	if (k==0) {
	    		alpha=alphastart;
	    	}
	    	if (k==epochs-1) {
	    		alpha=0.0001;
	    		r2=2;
	    	}


	    	// PASSO 2 DELL'ALGORITMO: Prende un vettore d'ingresso				
	    	for (int p=0; p<seq2mRNANumber; p++){
	    		if (k>0) ten=ten+1;
	    		if (k>0 & k<(epochs-1) & ten==updateStep){
	    			alpha=alphaend+(alphastart-alphaend)*((epochs-2)*seq2mRNANumber-pst*updateStep)/((epochs-2)*seq2mRNANumber);
	    			nRad=1+(size-1)*((epochs-2)*seq2mRNANumber-pst*updateStep)/((epochs-2)*seq2mRNANumber);
	    			r2=nRad*nRad+nRad*nRad;
	    			pst=pst+1;
	    			ten=0;
	    		}

	    		trainingSample = seq2mRNA.get(p);
	    		trv = new int[seqLength][4];
	    		changetobits(seqLength,trv,trainingSample);

			    //double thresholdHead = 4.9;
	    		double min = 99999;
	    		double min2 = 99999;
	    		min_i=-1;		
	    		min_j=-1;
	    		min_i2=-1;		
	    		min_j2=-1;


	    		// PASSO 3.1 DELL'ALGORITMO: Usa la distanza euclidea per trovare somiglianze fra il vettore d'ingresso e 
	    		//                           il vettore dei pesi di ogni singolo nodo della mappa
	    		for (int i=0; i<size; i++) {
	    			for (int j=0; j<size; j++) {
	    				distance=map[i][j].calculateNEDistance(trv);
	    				if (distance<min) {
	    					min2 = min;
	    					min_i2 = min_i; 
	    					min_j2 = min_j;
	    					min = distance;
	    					min_i = i; 
	    					min_j = j;
	    				}
	    				else if (distance<min2) {
	    					min2 = distance;
	    					min_i2 = i; 
	    					min_j2 = j;
	    				}
	    			}
	    		}
	    		// Se la distanza tra mRNA fragment e BMU e' minore di una soglia th, allora proietta mRNA fragment sulla mappa
	    		if (Math.sqrt(min) < thresholdHead){			    
	    			qe=qe+Math.sqrt(min);
	    			if (min_i-min_i2>1|min_i2-min_i>1|min_j-min_j2>1|min_j2-min_j>1) te=te+1;


	    			// PASSO 3.2 DELL'ALGORITMO: Individua il nodo a distanza minore (Best Matching Unit o BMU)					
	    			int bmu=min_j*size+min_i;
	    			int neuNro;

	    			String annot = annotSeqmRNA.get(seq2mRNAidx.get(p));
	    			mapmRNA[min_i][min_j].add(annot.split(" ")[0]+"_"+seq2mRNAsubidx.get(p));
	    			mapmRNA2[min_i][min_j].add(annot.split(" ")[0]);
	    		}
	    	} // end for seq2mRNANumber
	    } // end for epochs
// end ELABORAZIONE sui mRNA	

		
// ELABORAZIONE sui miRNA	
		for (int k=0; k<epochs; k++){
			qe=0;	
			te=0;
			int[] order = new int[seqmiRNANumber];
			//changeorder(order);
			
			if (k==0) {
				alpha=alphastart;
			}
			if (k==epochs-1) {
				alpha=0.0001;
				r2=2;
			}

			
			// PASSO 2 DELL'ALGORITMO: Prende un vettore d'ingresso				
			for (int p=0; p<seqmiRNANumber; p++){
				if (k>0) ten=ten+1;
				if (k>0 & k<(epochs-1) & ten==updateStep2){
					alpha=alphaend+(alphastart-alphaend)*((epochs-2)*seqmiRNANumber-pst*updateStep2)/((epochs-2)*seqmiRNANumber);
					nRad=1+(size-1)*((epochs-2)*seqmiRNANumber-pst*updateStep2)/((epochs-2)*seqmiRNANumber);
					r2=nRad*nRad+nRad*nRad;
					pst=pst+1;
					ten=0;
				}
				trainingSample = seqmiRNA.get(p);
				trv = new int[seqLength][4];
				changetobits(seqLength,trv,trainingSample);
				
				double min = 99999;
				double min2 = 99999;
				min_i=-1;		
				min_j=-1;
				min_i2=-1;		
				min_j2=-1;


				// PASSO 3.1 DELL'ALGORITMO: Usa la distanza euclidea per trovare somiglianze fra il vettore d'ingresso e 
				//                           il vettore dei pesi di ogni singolo nodo della mappa
				for (int i=0; i<size; i++) {
					for (int j=0; j<size; j++) {
						distance=map[i][j].calculateNEDistance(trv);
						if (distance<min) {
							min2 = min;
							min_i2 = min_i; 
							min_j2 = min_j;
							min = distance;
							min_i = i; 
							min_j = j;
						}
						else if (distance<min2) {
							min2 = distance;
							min_i2 = i; 
							min_j2 = j;
						}
					}
				}
				qe=qe+Math.sqrt(min);
				if (min_i-min_i2>1|min_i2-min_i>1|min_j-min_j2>1|min_j2-min_j>1) te=te+1;

			
				// PASSO 3.2 DELL'ALGORITMO: Individua il nodo a distanza minore (Best Matching Unit o BMU)					
				int bmu=min_j*size+min_i;
				int neuNro;
				
				//if(p == 8){
					//System.out.println(min_i + " , "+ min_j);
				//}
				
				String annot2 = annotSeqmiRNA.get(p);
				mapmiRNA[min_i][min_j].add(annot2.split(" ")[0]);						

			} // end for seqmiRNANumber
		} // end for epochs
// end ELABORAZIONE sui miRNA

		SOM result = new SOM();
		result.setMapmRNA(mapmRNA);
		result.setMapmRNA2(mapmRNA2);
		result.setMapmiRNA(mapmiRNA);

		return result;
	}

	private static void changetobits(int sL, int[][] trv, String sequence){
		int spit,plus;	
		char nt;
		int pts,j,p;
		
		sequence=sequence.toUpperCase();
		spit=sequence.length();

		if (spit<sL) {
			pts=sL-spit;
			plus=0;
		}
		else {
			pts=0;
			plus=spit-sL;
		}
		
		for (j=0;j<pts;j++){
			for (p=0;p<4;p++){
				trv[j][p]=0;
			}
		}
		
		for (j=pts;j<sL;j++){
			nt=sequence.charAt(j-pts+plus);
			switch (nt) {
			case 'A': trv[j][0]=1;
				break;
			case 'C': trv[j][1]=1;
				break;
			case 'G': trv[j][2]=1;
				break;
			case 'U': trv[j][3]=1;
				break;
			case 'T': trv[j][3]=1;
				break;
			default:  break;
			}
		}
	}
	
	private static void changeorder(int[] order){
		int seqN=order.length;
		
		int p;
		for (p=0; p<seqN; p++){
			order[p]=p;
		}
		int apu;
		int arpa;
		int laskuri=seqN;
		for (p=0; p<seqN; p++){
			arpa = (int)(laskuri*Math.random());
			apu=order[arpa];
			order[arpa]=order[laskuri-1];
			order[laskuri-1]=apu;
			laskuri=laskuri-1;
		}
	}
	
	private static void printAndSave (String s, PrintWriter pw) {
		//System.out.print(s);
		if (pw != null) pw.print(s);
	}
} 
