

import java.io.*;
import java.util.*;

// This class uses class Neuron and class LikenessType
public class SOMTrainingHead{

		public static String commonPath;
		public static Properties properties;
		
		//////// QUESTI VALORI LI STAMPO PER I TEST   ////////
		double alphastart;
		double alphaend;
		double qeMean= 0;
		double teMean= 0;
		double entropy=0;
		int mapSize;
		//////////////////////////////////////////////////////

		private ArrayList<String> seqmiRNA;  
		private ArrayList<String> annotSeqmiRNA;
		private Neuron[][] map;
		
		int epochs;

		File filemiRNA; // file di training dei miRNA

		ArrayList<String> EORList = new ArrayList<String>();
		ArrayList<String> EORList2 = new ArrayList<String>();
		PrintWriter outputFile;
		PrintWriter outputFile2;
		PrintWriter mapSOM;
		
		public Neuron[][] getMap() {
			return map;
		}
		public void setMap(Neuron[][] map) {
			this.map = map;
		}
		public ArrayList<String> getSeqmiRNA() {
			return seqmiRNA;
		}
		public void setSeqmiRNA(ArrayList<String> seqmiRNA) {
			this.seqmiRNA = seqmiRNA;
		}
		public ArrayList<String> getAnnotSeqmiRNA() {
			return annotSeqmiRNA;
		}
		public void setAnnotSeqmiRNA(ArrayList<String> annotSeqmiRNA) {
			this.annotSeqmiRNA = annotSeqmiRNA;
		}  
		
		
		public SOMTrainingHead(String _filemiRNA, int _mapSize, double _alphastart, double _alphaend, int _epochs){
			seqmiRNA = new ArrayList<String>();
			annotSeqmiRNA = new ArrayList<String>();
			filemiRNA = new File(_filemiRNA);
			mapSize = _mapSize;
			epochs = _epochs;
			alphastart= _alphastart;
			alphaend= _alphaend;
			//create a square matrix with dimensions mapSize x mapSize
			map = new Neuron[mapSize][mapSize];
			try {
				outputFile = new PrintWriter(new FileOutputStream(commonPath+"Output_SOMTrainingHead.txt"),true);
				outputFile2 = new PrintWriter(new FileOutputStream(commonPath+"Output_SOMTrainingHeadCSV.txt"),true);
				mapSOM = new PrintWriter(new FileOutputStream(commonPath+"SOMTrainingHead_"+mapSize+"_"+alphastart+"_"+alphaend+".txt"),true);
			} catch (FileNotFoundException e) {
				e.printStackTrace();
			}
		}

		/**
		 * This method is used for the source code path
		 * 
		 * @throws IOException
		 */
		public static void init() throws IOException {
			properties = new Properties();
			properties.load(new FileInputStream("prefs.properties"));
			commonPath = properties.getProperty("commonPath");
		}
		
		
		
		public static void main(String [] args) throws IOException{
			init();
			if (args.length != 5) {
				//System.out.println("Input example: java SOMTrainingHead miRNAfile 30 0.85 0.1 100");
				return;
			}

			String miRNAfile= args[0];
			String mapSize= args[1];
			String alphastart= args[2];
			String alphaend = args[3];
			String epochs= args[4];

			SOMTrainingHead som_mirna = new SOMTrainingHead(miRNAfile, Integer.valueOf(mapSize),  Double.valueOf(alphastart),  Double.valueOf(alphaend), Integer.parseInt(epochs) );

			try {
				som_mirna.computeSOMTrainingHead();
				som_mirna.printAndSave(som_mirna.EORList, som_mirna.EORList2, som_mirna.outputFile, som_mirna.outputFile2);
				som_mirna.printMap(som_mirna.mapSOM);
				//print  TEST output file
				som_mirna.printTestOutputFile();

				if (som_mirna.mapSOM != null)
					som_mirna.mapSOM.close(); 

				if (som_mirna.outputFile != null)
					som_mirna.outputFile.close(); 

				if (som_mirna.outputFile2 != null)
					som_mirna.outputFile2.close(); 

			} catch (RuntimeException e) {
				String msg = e.getMessage();
				if (msg == null) {
					throw e;
				}
				switch (msg) {
				case "File not found": 
				case "Errore nella lettura!": 
				case "Errore durante la chiusura del file!": 
					//System.out.println(msg);
					return;
				default: throw e;
				}
			}


		} // end main
		
		
	
	
	//Print output test file 
	private void printTestOutputFile() {
		try {
			//PrintWriter testoutput= new PrintWriter(new FileOutputStream(commonPath+"SOMtest4EANN2015.txt"),true);
			PrintWriter testoutput = new PrintWriter(new BufferedWriter(new FileWriter(commonPath+"SOMTrainingTest20160210.txt", true)));
			testoutput.write(mapSize+","+alphastart+","+alphaend+","+qeMean+","+teMean+","+entropy+"\n");

			testoutput.close();
			
		} catch (FileNotFoundException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
			//System.err.println(e);
		} catch (IOException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}	
			
		}
	
	
	
	public void computeSOMTrainingHead (){
		
		int seqLength = 8; // mirNA's SEED. It is a fixed value.
		String endOutputRow = null;
		String endOutputRow2 = null;

		if (!filemiRNA.exists()) {
			throw new RuntimeException("File not found");
		} else {
			readFileInput();		
		}
		
		int seqmiRNANumber = seqmiRNA.size();
		int units = mapSize*mapSize;
		double[][] Dist = new double[units][units]; // Matrice delle distanze (in termini di posizione) tra i neuroni nella griglia
	    InitDist (Dist, mapSize);
		

	    // PASSO 1 DEL TRAINING: Assegna ai vettori dei pesi valori casuali (INIZIALIZZAZIONE MAPPA)...
		initMap (seqLength);

		String trainingSample;
		double nRad = mapSize;	// neighbor radius
		double r2 = nRad*nRad+nRad*nRad;	
		double alpha=0.9;
		//alphastart = 0.9;	// training rate to start
		//alphaend = 0.05;	// 0.0001 //training rate in the end 

		double qe = 0;	// quantization error: Average distance between each input string and its best matching unit in the net
		qeMean= 0;
		double te = 0;	// topographic error: the proportion of all input strings for which the best and the second best matching units are not adjacent
		teMean= 0;
		double[] pN = new double[mapSize*mapSize]; //entropia della SOM: probabilita' delle unita' neurali
		int[][] trv;
		int ten = 0;
		int pst = 0;

		int updateStep = ((epochs-2)*seqmiRNANumber)/151;
		
		for (int k = 0; k<epochs; k++){
			qe = 0;	
			te = 0;
			int[] order = new int[seqmiRNANumber];
			changeOrder(order);
			
			if (k == 0) {
				alpha=alphastart;
			}
			if (k == epochs-1) {
				alpha=0.05;
				r2=2;
			}

			
			// PASSO 2 DEL TRAINING: Prende un vettore d'ingresso				
			for (int p=0; p<seqmiRNANumber; p++){
				if (k>0) ten=ten+1;
				//circa una volta ogni epoca aggiorna il valore di alpha e di r2
				if (k>0 & k<(epochs-1) & ten==updateStep){
					alpha=alphaend+(alphastart-alphaend)*((epochs-2)*seqmiRNANumber-pst*updateStep)/((epochs-2)*seqmiRNANumber);
					nRad=1+(mapSize-1)*((epochs-2)*seqmiRNANumber-pst*updateStep)/((epochs-2)*seqmiRNANumber);
					r2=nRad*nRad+nRad*nRad;
					pst=pst+1;
					ten=0;
//					System.out.println("epoch= "+k);
//					System.out.println("alpha= "+alpha);
//					System.out.println("r2= "+r2);
				}
				
				trainingSample = seqmiRNA.get(p);
				trv = new int[seqLength][4];
				changeToBits(seqLength,trv,trainingSample);


				// PASSO 3.1 DEL TRAININIG: Usa la distanza euclidea per trovare le somiglianze fra il vettore d'ingresso e il vettore dei pesi di ogni singolo nodo della mappa
				BMUType findBMU = findBMU(trv, mapSize);
				// qe e il quantization error medio di tutta la SOM.. 
				qe=qe+Math.sqrt(findBMU.getMin()); 
				//te e topographic error su base quadrata
				if (findBMU.getMin_i()-findBMU.getMin_i2()>1|findBMU.getMin2()-findBMU.getMin_i()>1|findBMU.getMin_j()-findBMU.getMin_j2()>1|findBMU.getMin_i2()-findBMU.getMin_j()>1) 
					te=te+1;


				// PASSO 3.2 DEL TRAINING: Individua il nodo a distanza minore (Best Matching Unit o BMU)					
				updateBMU(Dist, trv, alpha, r2, findBMU);
			
			} // end for seqmiRNANumber
			
		} // end for epochs
		
		//A mappa addestrata.... calcolo qe, te ed entropia
		for (int p=0; p<seqmiRNANumber; p++){
			double qErr = 0;
			trainingSample = seqmiRNA.get(p);
			trv = new int[seqLength][4];

			changeToBits(seqLength,trv,trainingSample);
			BMUType findBMU = findBMU(trv, mapSize);
			qErr = findBMU.getMin();
			qeMean=+qErr;
			teMean= +te;
			// 8 e la lunghezza del vettore 8-mers
			pN[findBMU.getMin_i()*8+findBMU.getMin_j()]=pN[findBMU.getMin_i()*8+findBMU.getMin_j()]+ 1-0;

			endOutputRow = annotSeqmiRNA.get(p)+": BMU = ["+String.valueOf(findBMU.getMin_i())+","+String.valueOf(findBMU.getMin_j())+"], QE = "+String.valueOf(qErr)+", TE = "+String.valueOf(te/seqmiRNANumber);
			endOutputRow2 = annotSeqmiRNA.get(p)+","+String.valueOf(findBMU.getMin_i())+","+String.valueOf(findBMU.getMin_j())+","+String.valueOf(qErr)+","+String.valueOf(te/seqmiRNANumber);
			EORList.add(endOutputRow);
			EORList2.add(endOutputRow2);
		}
		qeMean=(double)qeMean/seqmiRNANumber;
		teMean=(double)teMean/seqmiRNANumber;
		for (int b=0; b < pN.length; b++){
			pN[b] = pN[b] /seqmiRNANumber;
		}
		
		//calcolo l'entropia della rete	   	   
		for (int b=0; b< pN.length; b++) {
			if (pN[b] > 0.0)	entropy=entropy - pN[b] * Math.log(pN[b]);		
		}
			
		//System.out.println("Mean Quantization Error= "+qeMean);
		//System.out.println("Mean Topographic Error= "+teMean);
		//System.out.println("Entropy= "+entropy);
	} // end computeSOMTrainingHead

	
	private void readFileInput () {
		BufferedReader seqInput = null;
		String row = "";
  			try {
				seqInput = new BufferedReader(new FileReader(filemiRNA));
				while((row = seqInput.readLine())!=null){
					if ((row.indexOf(">")) == -1){
						seqmiRNA.add(row);
					}
					else annotSeqmiRNA.add(row);
				}
			} catch (IOException e1) {
				throw new RuntimeException("Errore nella lettura!", e1);
			}
			try {
				if (seqInput != null)
					seqInput.close();
			} catch (IOException e2) {
				throw new RuntimeException("Errore durante la chiusura del file!", e2);
			}
	}

	
	private void initMap (int seqLength){
		for (int i=0; i<mapSize; i++){
			for (int j=0; j<mapSize; j++) {
				map[i][j] = new Neuron();	
				map[i][j].setupWeights(seqLength,i,j);
			}
		}

	}

	
	private void InitDist(double[][] Dist, int mapSize) {
		int nr1,nr2;
		 for (int j1=0;j1<mapSize;j1++){
			for (int i1=0;i1<mapSize;i1++){	
				nr1=j1*mapSize+i1;	
				for (int j2=0;j2<mapSize;j2++){
					for (int i2=0;i2<mapSize;i2++){
						nr2=j2*mapSize+i2;
						if (nr1>nr2) 
							Dist[nr1][nr2]=Dist[nr2][nr1];
						else {
							Dist[nr1][nr2]=(j2-j1)*(j2-j1)+(i2-i1)*(i2-i1);
						}	
					}	
				}   
			}	
		 }	
		// TODO Auto-generated method stub
	}

	
	private BMUType findBMU(int[][] trv, int mapSize){
		double distance;
		BMUType findBMU = new BMUType();
		
		for (int i=0; i<mapSize; i++) {
			for (int j=0; j<mapSize; j++) {
				distance = map[i][j].calculateNEDistance(trv);
				if (distance<findBMU.getMin()) {
					findBMU.setMin2(findBMU.getMin());
					findBMU.setMin_i2(findBMU.getMin_i());
					findBMU.setMin_j2(findBMU.getMin_j());
					findBMU.setMin(distance);
					findBMU.setMin_i(i);
					findBMU.setMin_j(j);
				}
				else if (distance<findBMU.getMin2()) {
					findBMU.setMin2(distance);
					findBMU.setMin_i2(i); 
					findBMU.setMin_j2(j); 
				}
			}
		}

		return findBMU;
	}

	
	private void updateBMU(double[][] Dist, int[][] trv, double alpha, double r2, BMUType findBMU){
		int bmu = findBMU.getMin_j()*mapSize+findBMU.getMin_i();
		int neuNro;
		for (int i=0; i<mapSize; i++) {
			for (int j=0; j<mapSize; j++) {
				neuNro=j*mapSize+i;
				// PASSO 4 DEL TRAINING: Aggiorna i nodi del vicinato di BMU "tirandoli" piu vicino al vettore d'ingresso 
				if (Dist[bmu][neuNro]<=r2) map[i][j].updateWeights(trv,Dist[bmu][neuNro],alpha,r2);
			}
		}
		
	}
	
	
	private void printMap (PrintWriter pw) {
		if (pw != null) {
			for (int x=0; x<mapSize; x++) {
				for (int y=0; y<mapSize; y++) {
					map[x][y].printNodeWeight(pw,x,y);
				}
			}	

		}
	}
	

	private void printAndSave (ArrayList<String> list, ArrayList<String> list2, PrintWriter pw, PrintWriter pw2) {
		for (int i=0; i<list.size(); i++) 
			pw.println(list.get(i));					

		for (int i=0; i<list2.size(); i++) 
			pw2.println(list2.get(i));
	}

	
	private void changeToBits(int sL, int[][] trv, String sequence){
		int spit,plus;	
		char nt;
		int pts,j,p;
		
		sequence=sequence.toUpperCase();
		spit=sequence.length();

		if (spit<sL) {
			pts=sL-spit;
			plus=0;
		} else {
			pts=0;
			plus=spit-sL;
		}
		
		for (j=0;j<pts;j++){
			for (p=0;p<4;p++){
				trv[j][p]=0;
			}
		}
		
		for (j=pts;j<sL;j++){
			nt=sequence.charAt(j-pts+plus);
			switch (nt) {
			case 'A': trv[j][0]=1;
				break;
			case 'C': trv[j][1]=1;
				break;
			case 'G': trv[j][2]=1;
				break;
			case 'U': trv[j][3]=1;
				break;
			case 'T': trv[j][3]=1;
				break;
			default:  break;
			}
		}
	}
	
	
	private void changeOrder(int[] order){
		int seqN=order.length;

		int p;
		for (p=0; p<seqN; p++){
			order[p]=p;
		}
		int apu;
		int arpa;
		int laskuri=seqN;
		for (p=0; p<seqN; p++){
			arpa = (int)(laskuri*Math.random());
			apu=order[arpa];
			order[arpa]=order[laskuri-1];
			order[laskuri-1]=apu;
			laskuri=laskuri-1;
		}
	}
	
} // end class 